"""PXP: Python eXPression languate

PXP is a simple, python hosted expression language, intended to be used to calculate expressions
in code in a safe environment.
"""

from .compiler import Compiler
from .interpreter import Interpreter
from . import scope



"""This package contains classes and methods for executing compiled pxp code."""

from pxp.compiler import Compiler
from pxp.scope import ScopeStack, DictScope
from pxp.stdlib import global_scope
from .resolver import Resolver


class Interpreter(object):
  """The Interpreter class executes compiled pxp code."""
  def __init__(self, scope=None):
    """Initialize the Interpreter instance.

    If scope is present, it will be passed to the push method of the instance's ScopeStack.
    """
    self.scope = ScopeStack(global_scope)
    if scope:
      self.scope.push(scope)
    self.resolver = Resolver(self.scope)

  def interpret(self, source):
    """Compile and execute source code.

    source is pxp source text.
    """
    # Give the compiler all the non-global_scope scopes.
    compiler = Compiler(source, self.scope.scopes[1:])
    program = compiler.compile()
    return self.execute(program)

  @staticmethod
  def get_type(program):
    """Returns the return type of the return statement of the compiled program."""
    _, ptype, *_ = program
    return ptype

  def execute(self, program):
    """Execute compiled pxp instructions."""
    with self.scope.using(DictScope()):
      _, ptype, assignments, (_, return_exp) = program
      for assignment in assignments:
        self._assign(assignment)
      return self.resolver.resolve(return_exp)

  def _assign(self, assignment):
    """Handle an assignment instruction."""
    _, name, value = assignment
    self.scope.set_variable(name, value)



"""This module contains classes and methods related to interpreting and executing PXP programs."""

from ._interpreter import Interpreter, Resolver


def _main():
  """Handles command line interaction."""
  args = _parse_args()
  args.func(args)


def _parse_args():
  """Parses command line arguments."""
  import sys
  from argparse import ArgumentParser

  arg_parser = ArgumentParser(description="Compile or run pxp files")
  subparsers = arg_parser.add_subparsers()

  compile_parser = subparsers.add_parser("compile", description="Compile a pxp program")
  compile_parser.add_argument("in_filename",
                              nargs="?",
                              help="The file to compile. If not given you will be prompted to "
                                   "enter the program on the command line.")
  compile_parser.add_argument("out_filename",
                              help="The file to which to output the byte code. If the file has an "
                                   "extension other than .pxpc, that extension will be added to "
                                   "the filename that is given before output.")
  compile_parser.set_defaults(func=_compile)

  execute_parser = subparsers.add_parser("exec", description="Execute a pxp program")
  execute_parser.add_argument("filename",
                              nargs="?",
                              help="The file to execute. This can be a .pxp file, which will be "
                                   "compiled and executed, or a .pxpc file which will be executed "
                                   "directly. If not given you will be prompted to enter the "
                                   "program on the command line")
  execute_parser.set_defaults(func=_exec)

  args = arg_parser.parse_args()

  if hasattr(args, "func"):
    return args
  else:
    arg_parser.print_help()
    sys.exit(1)


def _compile(args):
  """Compile a pxp program."""
  import os
  import pickle
  from pxp.compiler import Compiler

  if args.in_filename:
    with open(args.in_filename) as rf:
      source = rf.read()
  else:
    source = _read_source()

  filename, ext = os.path.splitext(args.out_filename)
  if ext != ".pxpc":
    out_filename = args.out_filename + ".pxpc"
  else:
    out_filename = args.out_filename

  byte_code = Compiler(source).compile()
  with open(out_filename, "wb") as wf:
    pickle.dump(byte_code, wf)


def _exec(args):
  """Execute a pxp program."""
  import os
  from pxp.compiler import Compiler

  source = None
  byte_code = None

  if args.filename:
    _, ext = os.path.splitext(args.filename)
    if ext == ".pxpc":
      import pickle

      with open(args.filename, "rb") as rf:
        byte_code = pickle.load(rf)
    else:
      with open(args.filename, "r") as rf:
        source = rf.read()
  else:
    source = _read_source()

  if source:
    byte_code = Compiler(source).compile()

  print(Interpreter().execute(byte_code))


def _read_source(term="\\eof"):
  """Read a pxp program from the command line."""
  lines = []

  print("Enter the program then type {0} on its own line:".format(term))
  line = input("> ")
  while line != term:
    lines.append(line)
    line = input("> ")

  return "\n".join(lines)


if __name__ == "__main__":
  _main()



import string

from pyebnf.exception import ParserError
from pyebnf.operator import OperatorNode, OptreeNode, infix_to_optree
from pyebnf.primitive import NodeType

import pxp.util as util
from .ast import Assignment, Boolean, Branch, BranchOption, FunctionCall, Identifier, Number
from .ast import Program, ReturnStatement, String, Switch, SwitchCase, Variable, is_value
from .ast import from_value
from .exception import CompilerError, ScopeError
from .operator import operator_index
from .parser import Parser, TokenType
from .scope import DictScope, ScopeStack
from .stdlib import global_scope
from .stdlib.types import type_defaults


class Compiler(object):
  """This class generates an AST and interpreter code from PXP source code."""
  version = (0, 0, 1)

  def __init__(self, source, scope=None):
    """Initialize the compiler instance.

    If scope is present, it will be passed to the push method of the instance's ScopeStack.
    """
    self.source = _prepare_source(source)
    self.source_len = len(self.source)
    self.scope = ScopeStack(global_scope)
    if scope:
      self.scope.push(scope)
    self._source_newlines = None
    self._parse_tree = None
    self._abstract_syntax_tree = None
    self._externs = DictScope()
    self._resolving_variables = []

    # Add an empty scope to the end of the stack for the compiler to work with.
    self.scope.push(self._externs)

  def compile(self):
    """Returns the interpreter "byte code" from the AST generated from the source."""
    return self.abstract_syntax_tree.to_tuple()

  @property
  def parse_tree(self):
    """The parse tree represented by the source code."""
    if self._parse_tree is None:
      try:
        self._parse_tree = Parser().parse(self.source)
      except ParserError as pe:
        raise self._get_compiler_error(pe.position, pe.message)
      self._parse_tree.trim(_trim_predicate)
      self._parse_tree.flatten(_flatten_predicate)
    return self._parse_tree

  @property
  def abstract_syntax_tree(self):
    """The AST represented by the soruce code."""
    if self._abstract_syntax_tree is None:
      self._abstract_syntax_tree = self._generate_ast(self.parse_tree)[0]
    return self._abstract_syntax_tree

  @property
  def source_newlines(self):
    """The position of all the newlines in the source code. This is used for error reporting."""
    if self._source_newlines is None:
      self._source_newlines = util.get_newline_positions(self.source)
    return self._source_newlines

  def _get_node_coords(self, node):
    """Get the (line, char) coordinates of a node. Nodes have a position value which is their offset
    from the end of the source code. So their position from the beginning is first calculated, then
    that value is passed to _get_coords.
    """
    return self._get_coords(self.source_len - node.position)

  def _get_coords(self, pos):
    """Returns the (line, char) coordinates of an offset position in the source."""
    return util.get_line_and_char(self.source_newlines, pos)

  def _get_compiler_error(self, pos, message):
    """Build a CompilerError instance from a position and message.

    pos can be either a number, a 2-tuple of (line, char) or an object with a position property that
    represents its offset from the end of the source code (such as a NodeBase).
    """
    if isinstance(pos, int):
      coords = self._get_coords(pos)
    elif isinstance(pos, tuple):
      coords = pos
    else:
      coords = self._get_node_coords(pos)
    return CompilerError(coords, message)

  def _generate_ast(self, node, *, sibling=None):
    """Recursively builds the AST. This is done by handling nodes. A node handler returns a list
    of values to replace the node in the tree.
    """
    if not node.is_value:
      new_children = []
      prev_node = None
      for child in node.children:
        new_children.extend(self._generate_ast(child, sibling=prev_node))
        prev_node = child
      return self._handle_container_node(node, new_children, sibling=sibling)
    else:
      return self._handle_value_node(node, sibling=sibling)

  def _handle_container_node(self, node, new_children, *, sibling=None):
    """A switching method that delegates to another method based on the type of the containt node.
    """
    if node.is_type(TokenType.program):
      return self._handle_program(node, new_children)
    elif node.is_type(TokenType.assignment):
      return self._handle_assignment(node, new_children)
    elif node.is_type(TokenType.expression):
      return self._handle_expression(node, new_children)
    elif node.is_type(TokenType.subexpression):
      return self._handle_subexpression(node, new_children)
    elif node.is_type(TokenType.function_call):
      return self._handle_function_call(node, new_children)
    elif node.is_type(TokenType.return_statement):
      return self._handle_return_statement(node, new_children)
    elif node.is_type((TokenType.branch_if, TokenType.branch_elif)):
      return self._handle_branch_option(node, new_children)
    elif node.is_type(TokenType.branch_else):
      return self._handle_branch_default(node, new_children)
    elif node.is_type(TokenType.branch):
      return self._handle_branch(node, new_children)
    elif node.is_type(TokenType.switch_subject):
      return self._handle_switch_subject(node, new_children)
    elif node.is_type(TokenType.switch_case):
      return self._handle_switch_case(node, new_children)
    elif node.is_type(TokenType.switch_default):
      return self._handle_switch_default(node, new_children)
    elif node.is_type(TokenType.switch):
      return self._handle_switch(node, new_children)
    else:
      raise Exception("Unhandled node type {0}".format(node.node_type))

  def _handle_value_node(self, node, *, sibling=None):
    """A switching method that delegates to another method based on the type of the value node."""
    if node.is_type(TokenType.comment):
      return self._ignore_node()
    elif node.is_type(TokenType.variable):
      return self._handle_variable(node)
    elif node.is_type(TokenType.string):
      return self._handle_string(node)
    elif node.is_type(TokenType.number):
      return self._handle_number(node)
    elif node.is_type(TokenType.operator):
      return self._handle_operator(node, sibling=sibling)
    elif node.is_type(TokenType.function_name):
      return self._handle_function_name(node)
    elif node.is_type(TokenType.directive):
      return self._handle_directive(node)
    else:
      raise Exception("Unhandled node type {0}".format(node.node_type))

  def _ignore_node(self):
    """Ignores a node by returning an empty list."""
    return []

  def _handle_program(self, node, new_children):
    """Handles AST processing for program PT nodes."""
    return [Program(node, new_children)]

  def _handle_assignment(self, node, new_children):
    """Handles AST processing for assignment PT nodes."""
    assignee, assignment = new_children
    anode = Assignment(node, assignee, assignment)
    self.scope.set_variable(assignee.value, assignment)
    self.scope.set_variable(assignee.value + ".__assign__", anode)
    return [anode]

  def _handle_variable(self, node):
    """Handles AST processing for variable PT nodes."""
    if node.svalue in ("true", "false"):
      return [Boolean.from_parse_node(node, self._get_node_coords(node))]
    else:
      return [Variable.from_parse_node(node, self._get_node_coords(node))]

  def _handle_string(self, node):
    """Handles AST processing for string PT nodes."""
    return [String.from_parse_node(node, self._get_node_coords(node))]

  def _handle_number(self, node):
    """Handles AST processing for number PT nodes."""
    return [Number.from_parse_node(node, self._get_node_coords(node))]

  def _handle_operator(self, node, *, sibling=None):
    """Handles AST process for operator PT nodes."""
    # If an operator is the first item in an expression, or follows an operator, then it is
    # considered to be a unary operator and we must prefix it's symbol with 'unary' to disambiguate
    # it from binary oeprators that use the same symbol. E.G. -5 and 3 - 5 and 3 - -5.
    if sibling is None or sibling.is_type(TokenType.operator):
      operator_name = "unary{0}".format(node.svalue)
    else:
      operator_name = node.svalue

    return [OperatorNode(operator_index[operator_name], self._get_node_coords(node))]

  def _handle_expression(self, node, new_children):
    """Handles AST process for expression PT nodes."""
    return [infix_to_optree(new_children)]

  def _handle_subexpression(self, node, new_children):
    """Handles AST process for subexpression PT nodes."""
    return new_children

  def _handle_function_name(self, node):
    """Handles AST process for function name PT nodes."""
    return [Identifier.from_parse_node(node, self._get_node_coords(node))]

  def _handle_function_call(self, node, new_children):
    """Handles AST process for function call PT nodes."""
    name, *args = new_children
    return [FunctionCall(self._get_node_coords(node), name, args)]

  def _handle_return_statement(self, node, new_children):
    """Handles AST process for return statement PT nodes."""
    optree_type, optree = self._resolve_optree(new_children[0])
    return [ReturnStatement(node, optree, optree_type)]

  def _handle_directive(self, node):
    """Handles AST process for directive PT nodes."""
    dir_parts = node.value.strip().split(maxsplit=1)
    command = dir_parts[0]
    if command == "#extern":
      if len(dir_parts) == 2:
        arg_parts = dir_parts[1].split(maxsplit=1)
        if len(arg_parts) == 2:
          etype, ename = arg_parts
          if etype not in type_defaults:
            raise self._get_compiler_error(node, "Unknown type {0} for {1} directive".format(etype, command))
          else:
            self._externs.set_variable(ename, type_defaults[etype])
        else:
          raise self._get_compiler_error(node, "Directive {0} expects exactly 2 arguments".format(command))
      else:
        raise self._get_compiler_error(node, "Directive {0} expects arguments".format(command))
    else:
      raise self._get_compiler_error(node, "Unknown directive: {0}".format(command))
    return []

  def _handle_branch_option(self, node, new_children):
    predicate, result = new_children
    return [BranchOption(self._get_node_coords(node), result, predicate)]

  def _handle_branch_default(self, node, new_children):
    return [BranchOption(self._get_node_coords(node), new_children[0], None)]

  def _handle_branch(self, node, new_children):
    return [Branch(self._get_node_coords(node), new_children)]

  def _handle_switch_subject(self, node, new_children):
    return new_children

  def _handle_switch_case(self, node, new_children):
    comparison, result = new_children
    return [SwitchCase(self._get_node_coords(node), result, comparison)]

  def _handle_switch_default(self, node, new_children):
    return [SwitchCase(self._get_node_coords(node), new_children[0], None)]

  def _handle_switch(self, node, new_children):
    subject, *cases = new_children
    return [Switch(self._get_node_coords(node), subject, cases)]

  def _resolve(self, obj):
    """A switch method for resolving AST expression nodes."""
    if isinstance(obj, OptreeNode):
      return self._resolve_optree(obj)
    elif isinstance(obj, Variable):
      return self._resolve_variable(obj)
    elif isinstance(obj, FunctionCall):
      return self._resolve_function_call(obj)
    elif isinstance(obj, Branch):
      return self._resolve_branch(obj)
    elif isinstance(obj, BranchOption):
      return self._resolve_branch_option(obj)
    elif isinstance(obj, Switch):
      return self._resolve_switch(obj)
    elif is_value(obj):
      astn = from_value(obj)
      return (astn.type, astn)
    else:
      raise Exception("Unhandled type {0}".format(obj.__class__.__name__))

  def _resolve_optree(self, optree):
    """Resolves AST optree nodes."""
    new_operands = []
    types = []

    for operand in optree.operands:
      otype, newop = self._resolve(operand)
      new_operands.append(newop)
      types.append(otype)

    # operator is None -> Identity
    opnode = optree.opnode
    if opnode is None:
      return (types[0], new_operands[0])
    else:
      op_fn_name = "operator{0}".format(opnode.operator.symbol)
      fn_sig = (op_fn_name, ) + tuple(types)
      try:
        fn = self.scope.get_function(fn_sig)
      except ScopeError as ex:
        raise self._get_compiler_error(opnode.position, str(ex))
      return (fn.return_type, FunctionCall(opnode.position, Identifier(op_fn_name), new_operands, fn))

  def _resolve_variable(self, variable):
    """Resolves AST variable nodes."""
    name = variable.value
    if name in self._resolving_variables:
      raise self._get_compiler_error(variable.coords, "Circular reference: {0} -> {1}".format(" -> ".join(self._resolving_variables), name))

    try:
      value = self.scope.get_variable(name)
    except ScopeError as ex:
      raise self._get_compiler_error(variable.coords, str(ex))

    self._resolving_variables.append(name)

    if variable.type is None:
      otype, newop = self._resolve(value)
      variable.type = otype

      if not self.scope.is_constant(name):
        self.scope.set_variable(name, newop)

        assign_var_name= "{0}.__assign__".format(name)
        if self.scope.contains_variable(assign_var_name):
          assn_node = self.scope.get_variable(assign_var_name)
          assn_node.assignment = newop
          assn_node.is_used = True

    self._resolving_variables.pop()

    return (variable.type, variable)

  def _resolve_function_call(self, function_call):
    """Resolves AST function call nodes."""
    if function_call.function is None:
      types, new_args = zip(*[self._resolve(arg) for arg in function_call.args])
      fn_sig = (function_call.name.value, ) + tuple(types)
      try:
        fn = self.scope.get_function(fn_sig)
      except ScopeError as ex:
        raise self._get_compiler_error(function_call.name.coords, str(ex))
      function_call.args = new_args
      function_call.function = fn

    return (function_call.function.return_type, function_call)

  def _resolve_branch(self, branch):
    for option in branch.options:
      self._resolve_branch_option(option)
    btype = branch.options[0].type
    if not all(o.type == btype for o in branch.options):
      raise self._get_compiler_error(branch.coords, "Not all branches have the same return type")
    return (branch.type, branch)

  def _resolve_branch_option(self, option):
    rtype, res = self._resolve(option.result)
    option.result = res
    if option.predicate:
      ptype, pred = self._resolve(option.predicate)
      if ptype != "Boolean":
        raise self._get_compiler_error(option.coords, "Branch predicate is not Boolean type")
      option.predicate = pred
    return (rtype, option)

  def _resolve_switch(self, switch):
    stype, subj = self._resolve(switch.subject)
    switch.subject = subj
    for case in switch.cases:
      self._resolve_switch_case(case)
    rtype = switch.cases[0].type
    for case in switch.cases:
      if case.comparison:
        if case.comparison.type != stype:
          raise self._get_compiler_error(case.coords, "Switch case predicate type doesn't match subject type")
      if case.result.type != rtype:
        raise self._get_compiler_error(case.coords, "Switch case return type doesn't match others")
    return (rtype, switch)

  def _resolve_switch_case(self, case):
    rtype, res = self._resolve(case.result)
    case.result = res
    if case.comparison:
      ctype, comp = self._resolve(case.comparison)
      case.comparison = comp
    return (rtype, case)


def _flatten_predicate(c, p):
  """Predicate passed to NodeBase for flattening children."""
  if c.is_type(TokenType.expression) and p.node_type == c.node_type:
    return True
  if c.is_type(TokenType.function_args) and p.is_type({TokenType.function_args, TokenType.function_call}):
    return True
  if isinstance(c.node_type, NodeType) and not c.is_type(NodeType.terminal):
    return True
  return False


def _trim_predicate(c):
  """Predicate passed to NodeBase for trimming children."""
  return c.is_type(NodeType.terminal)


def _prepare_source(source, eof_marker="\\eof"):
  """Given raw source, prepares it to make sure it can be handled by the parser. Specifically by
  adding the required EOF marker.
  """
  source = source.rstrip(string.whitespace)
  if not source.endswith(eof_marker):
    return source + "\n{0}".format(eof_marker)
  return source



"""This module contains definitions for the operators that are recognized by PXP programs."""

from pyebnf.operator import Association, Operator


# Arithmetic
op_add = Operator("+", 5)
op_subtract = Operator("-", 5)
op_multiply = Operator("*", 6)
op_divide = Operator("/", 6)
op_modulus = Operator("%", 6)
op_exponentiate = Operator("^", 8, Association.right)
op_negate = Operator("unary-", 7, Association.right, cardinality=1)

# Comparison
op_cmp_equal = Operator("=", 1)
op_cmp_not_equal = Operator("!=", 1)
op_cmp_greater_than_or_equal = Operator(">=", 1)
op_cmp_less_than_or_equal = Operator("<=", 1)
op_cmp_greater_than = Operator(">", 1)
op_cmp_less_than = Operator("<", 1)

# Null
op_null_coalesce = Operator("?", 7)

# Logic
op_logical_not = Operator("unary!", 4, Association.right, cardinality=1)
op_logical_or = Operator("|", 2)
op_logical_and = Operator("&", 3)

operators = [
  op_add,
  op_subtract,
  op_multiply,
  op_divide,
  op_modulus,
  op_exponentiate,
  op_negate,
  op_cmp_equal,
  op_cmp_not_equal,
  op_cmp_greater_than_or_equal,
  op_cmp_less_than_or_equal,
  op_cmp_greater_than,
  op_cmp_less_than,
  op_null_coalesce,
  op_logical_not,
  op_logical_or,
  op_logical_and
]

operator_index = {op.symbol: op for op in operators}



"""This module contains the Resolver class which is used by pxp functions to resolve their
arguments.
"""

from decimal import Decimal

from pxp.exception import FunctionError, RuntimeError, ScopeError


class Resolver(object):
  """The Resolver class is used by pxp functions to resolve their arguments. Arguments may be
  values, variables or expressions. The resolver class handles each and returns the actual value
  represented by the argument."""

  def __init__(self, scope):
    """Initialize the Resolver instance.

    scope is a ScopeBase instance that the resolver can use to look up variables and functions.
    """
    self.scope = scope

  def resolve(self, arg, *, none_ok=False):
    """Resolve an argument.

    If arg is a tuple, it is expected to be a raw compiled argument. The first value in the tuple
    is a string representing the type, one of: {fn, var, val} and the remaining items are the
    details of that value. If arg is not a tuple it is a real value that is passed directly back
    to the caller.
    """
    if isinstance(arg, tuple):
      arg_type = arg[0]
      # NOTE: These are 'magical' string values and aren't strongly tied with the AST module.
      if arg_type == "fn":
        val = self._resolve_function(arg)
      elif arg_type == "var":
        val = self._resolve_variable(arg, none_ok=none_ok)
      elif arg_type == "val":
        val = self._resolve_value(arg)
      elif arg_type == "branch":
        val = self._resolve_branch(arg)
      elif arg_type == "switch":
        val = self._resolve_switch(arg)
      else:
        raise Exception("Unknown arg type: {0}".format(arg_type))
    else:
      val = arg

    if val is None and not none_ok:
      raise FunctionError("Unexpected null value")

    return val

  def resolve_all(self, *args):
    """Resolves multiple arguments at once and returns their values as a tuple in the order they
    were specified.
    """
    return (self.resolve(arg) for arg in args)

  def _resolve_function(self, arg):
    """Resolves a function instruction."""
    _, signature, fn_args, position = arg
    fn = self.scope.get_function(signature)
    try:
      return fn(self, *fn_args)
    except FunctionError as ex:
      raise RuntimeError(position, str(ex))
    except TypeError as ex:
      raise RuntimeError(position, str(ex))

  def _resolve_variable(self, arg, *, none_ok=False):
    """Resolves a variable instruction."""
    _, name, position = arg
    try:
      raw_value = self.scope.get_variable(name)
    except ScopeError as ex:
      raise RuntimeError(position, str(ex))

    value = self.resolve(raw_value, none_ok=none_ok)
    if value != raw_value:
      self.scope.overwrite_variable(name, value)
    return value

  def _resolve_value(self, arg):
    """Resolves a value instruction."""
    _, vtype, vstr = arg
    if vtype == "Number":
      return Decimal(vstr)
    elif vtype == "String":
      return vstr
    elif vtype == "Boolean":
      return vstr == "True"
    else:
      raise Exception("Unknown value type: {0}".format(vtype))

  def _resolve_branch(self, arg):
    _, options, position = arg
    for option in options:
      passes, val = self._resolve_branch_option(option)
      if passes:
        return val
    raise RuntimeError(position, "Non-exhaustive branch")

  def _resolve_branch_option(self, arg):
    _, result, pred = arg
    if pred is None or self.resolve(pred) == True:
      return (True, self.resolve(result))
    return (False, None)

  def _resolve_switch(self, arg):
    _, subj, cases, position = arg
    subject = self.resolve(subj)
    for case in cases:
      _, result, comparison = case
      if comparison is None or self.resolve(comparison) == subject:
        return self.resolve(result)
    raise RuntimeError(position, "Non-exhaustive switch")



"""This module contains utility functions used by the PXP internals."""

import math


def get_newline_positions(text):
  """Returns a list of the positions in the text where all new lines occur. This is used by
  get_line_and_char to efficiently find coordinates represented by offset positions.
  """
  pos = []
  for i, c in enumerate(text):
    if c == "\n":
      pos.append(i)
  return pos


def get_line_and_char(newline_positions, position):
  """Given a list of newline positions, and an offset from the start of the source code
  that newline_positions was pulled from, return a 2-tuple of (line, char) coordinates.
  """
  if newline_positions:
    for line_no, nl_pos in enumerate(newline_positions):
      if nl_pos >= position:
        if line_no == 0:
          return (line_no, position)
        else:
          return (line_no, position - newline_positions[line_no - 1] - 1)
    return (line_no + 1, position - newline_positions[-1] - 1)
  else:
    return (0, position)


def point_to_source(source, position, fmt=(2, True, "~~~~~", "^")):
  """Point to a position in source code.

  source is the text we're pointing in.
  position is a 2-tuple of (line_number, character_number) to point to.
  fmt is a 4-tuple of formatting parameters, they are:
    name               default  description
    ----               -------  -----------
    surrounding_lines  2        the number of lines above and below the target line to print
    show_line_numbers  True     if true line numbers will be generated for the output_lines
    tail_body          "~~~~~"  the body of the tail
    pointer_char       "^"      the character that will point to the position
  """

  surrounding_lines, show_line_numbers, tail_body, pointer_char = fmt

  line_no, char_no = position

  lines = source.split("\n")
  line = lines[line_no]

  if char_no >= len(tail_body):
    tail = " " * (char_no - len(tail_body)) + tail_body + pointer_char
  else:
    tail = " " * char_no + pointer_char + tail_body

  if show_line_numbers:
    line_no_width = int(math.ceil(math.log10(max(1, line_no + surrounding_lines))) + 1)
    line_fmt = "{0:" + str(line_no_width) + "}: {1}"
  else:
    line_fmt = "{1}"

  pivot = line_no + 1
  output_lines = [(pivot, line), ("", tail)]
  for i in range(surrounding_lines):
    upper_ofst = i + 1
    upper_idx = line_no + upper_ofst
    lower_ofst = -upper_ofst
    lower_idx = line_no + lower_ofst

    if lower_idx >= 0:
      output_lines.insert(0, (pivot + lower_ofst, lines[lower_idx]))
    if upper_idx < len(lines):
      output_lines.append((pivot + upper_ofst, lines[upper_idx]))

  return "\n".join(line_fmt.format(n, c) for n, c in output_lines)



"""This package contains the implementaion of pxp standard library methods and constants."""

from pxp.function import FunctionList
from pxp.scope import DictScope

from pxp.stdlib.general import general_functions
from pxp.stdlib.operator import operator_functions
from pxp.stdlib.math import math_functions, math_constants
from pxp.stdlib.string import string_functions


_function_list = FunctionList()
_function_list.merge(general_functions,
                     operator_functions,
                     math_functions,
                     string_functions)


global_scope = DictScope(None, math_constants, _function_list)



"""This module contains classes and methods for working with the PXP Abstract Syntax Tree (AST)."""

from decimal import Decimal

from pyebnf.operator import OptreeNode


def pprint(node, depth=0, indent="    "):
  """Print out a PXP AST in a "pretty" way."""
  spacing = indent * depth

  if isinstance(node, ASTLeaf):
    print("{0}{1}".format(spacing, node))
  elif isinstance(node, OptreeNode):
    opnode = node.opnode
    if opnode:
      op = opnode.operator
      print("{0}Operator({1}, {2})".format(spacing, op.symbol, node.type))
    else:
      print("{0}Operator(None -> Identity)".format(spacing))
    for operand in node.operands:
      pprint(operand, depth + 1, indent)
  else:
    print("{0}{1}:".format(spacing, node.__class__.__name__))
    for child in node.children:
      pprint(child, depth + 1, indent)


class ASTNodeBase(object):
  """Base class for all AST Nodes."""
  def __init__(self, coords):
    """Construct an ASTNodeBase instance."""
    self.coords = coords

  def to_tuple(self):
    """Returns a tuple representing the node that can be understood by the PXP interpreter."""
    raise NotImplementedError()


class ASTNode(ASTNodeBase):
  """Base class for AST grouping nodes. A grouping node has children, as compared to leaf (value)
  nodes, which contains just a value.
  """
  def __init__(self, coords=None):
    """Construct an ASTNode instance."""
    super().__init__(coords)

  @property
  def children(self):
    """The children of the Node."""
    raise StopIteration()


class ASTLeaf(ASTNodeBase):
  """Base class for AST leaf nodes. Leaf nodes do not have children of their own, but rather
  represent leaves of the AST.
  """
  def __init__(self, value, type, coords=None):
    """Construct an ASTNode instance."""
    super().__init__(coords)
    self.value = value
    self.type = type

  def __repr__(self):
    """Returns a representation of the Node, including its type and value."""
    return "{0}({1})".format(self.__class__.__name__, self.value)

  def __str__(self):
    """Returns a string representation of the Node."""
    return repr(self)

  def to_tuple(self):
    """Returns a tuple representing the node that can be understood by the PXP interpreter."""
    return ("val", self.type, str(self.value))


class Program(ASTNode):
  """The Program (root) node of the AST."""
  def __init__(self, coords, statements):
    """Construct a Program node."""
    super().__init__(coords)
    self.statements = statements

  @property
  def children(self):
    """The statements that make up the program."""
    yield from self.statements

  def to_tuple(self):
    """Returns a tuple representing the node that can be understood by the PXP interpreter."""
    *assignments, ret_statement = self.statements
    return ("program",
            ret_statement.type,
            tuple(a.to_tuple() for a in assignments if a.is_used),
            ret_statement.to_tuple())


class Assignment(ASTNode):
  """An Assignment node."""
  def __init__(self, coords, assignee, assignment):
    """Construct an Assignment node."""
    super().__init__(coords)
    self.assignee = assignee
    self.assignment = assignment
    self.is_used = False

  @property
  def children(self):
    """The assignee and assignment value representing the Assignment node."""
    yield self.assignee
    yield self.assignment

  def to_tuple(self):
    """Returns a tuple representing the node that can be understood by the PXP interpreter."""
    return ("assign", self.assignee.value, self.assignment.to_tuple())


class ReturnStatement(ASTNode):
  """The ReturnStatement of the program."""
  def __init__(self, coords, expression, type):
    """Construct a ReturnStatement node."""
    super().__init__(coords)
    self.expression = expression
    self.type = type

  @property
  def children(self):
    """The expression to be evaluated for return."""
    yield self.expression

  def to_tuple(self):
    """Returns a tuple representing the node that can be understood by the PXP interpreter."""
    return ("return", self.expression.to_tuple())


class FunctionCall(ASTNode):
  """A node representing a function call."""
  def __init__(self, coords, name, args, function=None):
    """Construct a FunctionCall instance."""
    super().__init__(coords)
    self.name = name
    self.args = args
    self.function = function

  @property
  def children(self):
    """The name and args of the function call."""
    yield self.name
    yield from self.args

  @property
  def type(self):
    """The return type of the function call."""
    return self.function.return_type

  def to_tuple(self):
    """Returns a tuple representing the node that can be understood by the PXP interpreter."""
    return ("fn", self.function.signatures[0], tuple(a.to_tuple() for a in self.args), self.coords)


class Branch(ASTNode):
  def __init__(self, coords, options):
    super().__init__(coords)
    self.options = options

  @property
  def children(self):
    yield from self.options

  @property
  def type(self):
    return self.options[0].type

  def to_tuple(self):
    return ("branch", tuple(o.to_tuple() for o in self.options), self.coords)


class BranchOption(ASTNode):
  def __init__(self, coords, result, predicate=None):
    super().__init__(coords)
    self.result = result
    self.predicate = predicate

  @property
  def children(self):
    yield self.result
    if self.predicate:
      yield self.predicate

  @property
  def type(self):
    return self.result.type

  def to_tuple(self):
    return ("branch_option",
            self.result.to_tuple(),
            self.predicate.to_tuple() if self.predicate else None)


class Switch(ASTNode):
  def __init__(self, coords, subject, cases):
    super().__init__(coords)
    self.subject = subject
    self.cases = cases

  @property
  def children(self):
    yield self.subject
    yield from self.cases

  @property
  def type(self):
    return self.cases[0].type

  def to_tuple(self):
    return ("switch",
            self.subject.to_tuple(),
            tuple(c.to_tuple() for c in self.cases),
            self.coords)


class SwitchCase(ASTNode):
  def __init__(self, coords, result, comparison=None):
    super().__init__(coords)
    self.result = result
    self.comparison = comparison

  @property
  def children(self):
    yield self.result
    if self.comparison:
      yield self.comparison

  @property
  def type(self):
    return self.result.type

  def to_tuple(self):
    return ("switch_case",
            self.result.to_tuple(),
            self.comparison.to_tuple() if self.comparison else None)


class Number(ASTLeaf):
  """A literal numeric value node."""
  def __init__(self, value, coords=None):
    """Construct the Number instance."""
    super().__init__(Decimal(value), "Number", coords)

  @classmethod
  def from_parse_node(cls, parse_node, coords):
    """Construct a Number instance from a parse tree node."""
    return cls(parse_node.svalue, coords)


class Boolean(ASTLeaf):
  """A literal boolean value node."""
  def __init__(self, value, coords=None):
    """Construct the Boolean instance."""
    super().__init__(value, "Boolean", coords)

  @classmethod
  def from_parse_node(cls, parse_node, coords):
    """Construct a Boolean instance from a parse tree node."""
    return cls(parse_node.svalue == "true", coords)


class String(ASTLeaf):
  """A literal string value node."""
  def __init__(self, value, coords=None):
    """Construct the String instance."""
    super().__init__(value, "String", coords)

  @classmethod
  def from_parse_node(cls, parse_node, coords):
    """Construct a String instance from a parse tree node."""
    return cls(parse_node.svalue[1:-1], coords)


class Identifier(ASTLeaf):
  """An identifier node."""
  def __init__(self, value, type=None, coords=None):
    """Construct the Identifier instance."""
    super().__init__(value, type, coords)

  @classmethod
  def from_parse_node(cls, parse_node, coords):
    """Construct an Identifier instance from a parse tree node."""
    if parse_node.svalue.startswith("["):
      value = parse_node.svalue[1:-1]
    else:
      value = parse_node.svalue

    return cls(value, None, coords)


class Variable(Identifier):
  """A variable instance, which is a special type of identifier."""
  pass

  def to_tuple(self):
    """Returns a tuple representing the node that can be understood by the PXP interpreter."""
    return ("var", self.value, self.coords)


def is_value(obj):
  """Returns True if the object is either an ASTLeaf or a python object that is considered a raw
  value in PXP."""
  return isinstance(obj, (ASTLeaf, Decimal, str, bool))


def from_value(obj):
  """If obj is an ASTLeaf, returns the leaf. Otherwise either a Number, String or Boolean ASTLeaf
  is returned based on the Python type of the object.
  """
  if isinstance(obj, ASTLeaf):
    return obj
  elif isinstance(obj, Decimal):
    return Number(obj)
  elif isinstance(obj, str):
    return String(obj)
  elif isinstance(obj, bool):
    return Boolean(obj)
  else:
    raise Exception("Unhandled value conversion type: {0}".format(obj.__class__.__name__))



import math
from decimal import Decimal

from pxp.exception import FunctionError
from pxp.function import FunctionArg, FunctionList, InjectedFunction
from pxp.stdlib.types import number_t, boolean_t


def math_abs(resolver, value):
  """Returns the absolute value of value."""
  val = resolver.resolve(value)
  return val if val >= 0 else -val


def math_ceil(resolver, value):
  """Returns value if value is a whole number, otherwise the next largest whole number."""
  val = resolver.resolve(value)
  return Decimal(math.ceil(val))


def math_cos(resolver, value):
  """Returns the cosine of value. Value must be in radians."""
  val = resolver.resolve(value)
  return Decimal(math.cos(val))


def math_degrees(resolver, value):
  """Converts a radians value to degrees."""
  val = resolver.resolve(value)
  return Decimal(math.degrees(val))


def math_floor(resolver, value):
  """Returns value if value is a whole number, otherwise the next smallest whole number."""
  val = resolver.resolve(value)
  return Decimal(math.floor(val))


def math_log(resolver, value, base):
  """Returns the log of value. If not specified, the log is a natural log with base e."""
  bval = resolver.resolve(base)

  if bval <= Decimal(0):
    raise FunctionError("Invalid log base")

  val = resolver.resolve(value)
  return Decimal(math.log(val, bval))


def math_log10(resolver, value):
  """Returns the log base 10 of value."""
  return math_log(resolver, value, Decimal(10))


def math_log2(resolver, value):
  """Returns the log base 2 of value."""
  return math_log(resolver, value, Decimal(2))


def math_pow(resolver, value, exp):
  """Returns value raised to exp."""
  val = resolver.resolve(value)
  xval = resolver.resolve(exp)
  return Decimal(math.pow(val, xval))


def math_radians(resolver, value):
  """Converts a degrees value to radians."""
  val = resolver.resolve(value)
  return Decimal(math.radians(val))


def math_root(resolver, value, root):
  """Returns the nth root of value."""
  val = resolver.resolve(value)
  rval = resolver.resolve(root)
  return Decimal(math.pow(val, Decimal(1) / rval))


def math_round(resolver, value, ndigits):
  """Rounds value to the nearest nth digit.

  If ndigits is not specified then value is rounded to the nearest whole number.
  """
  val = resolver.resolve(value)
  dval = resolver.resolve(ndigits)
  return Decimal(round(val, int(dval)))


def math_sin(resolver, value):
  """Returns the sine of value. Value must be in radians."""
  val = resolver.resolve(value)
  return Decimal(math.sin(val))


def math_sqrt(resolver, value):
  """Returns the square root of value."""
  return math_root(resolver, value, Decimal(2))


def math_tan(resolver, value):
  """Returns the tanget of value. Value must be in radians."""
  val = resolver.resolve(value)
  return Decimal(math.tan(val))



math_functions = FunctionList((
  InjectedFunction("math.abs", (FunctionArg(number_t, "value"), ), number_t, math_abs),
  InjectedFunction("math.ceil", (FunctionArg(number_t, "value"), ), number_t, math_ceil),
  InjectedFunction("math.cos", (FunctionArg(number_t, "value"), ), number_t, math_cos),
  InjectedFunction("math.degrees", (FunctionArg(number_t, "value"), ), number_t, math_degrees),
  InjectedFunction("math.floor", (FunctionArg(number_t, "value"), ), number_t, math_floor),
  InjectedFunction("math.log", (FunctionArg(number_t, "value"), FunctionArg(number_t, "base", Decimal(math.e))), number_t, math_log),
  InjectedFunction("math.log10", (FunctionArg(number_t, "value"), ), number_t, math_log10),
  InjectedFunction("math.log2", (FunctionArg(number_t, "value"), ), number_t, math_log2),
  InjectedFunction("math.pow", (FunctionArg(number_t, "value"), FunctionArg(number_t, "exp")), number_t, math_pow),
  InjectedFunction("math.radians", (FunctionArg(number_t, "value"), ), number_t, math_radians),
  InjectedFunction("math.root", (FunctionArg(number_t, "value"), FunctionArg(number_t, "root")), number_t, math_root),
  InjectedFunction("math.round", (FunctionArg(number_t, "value"), FunctionArg(number_t, "ndigits", Decimal(0))), number_t, math_round),
  InjectedFunction("math.sin", (FunctionArg(number_t, "value"), ), number_t, math_sin),
  InjectedFunction("math.sqrt", (FunctionArg(number_t, "value"), ), number_t, math_sqrt),
  InjectedFunction("math.tan", (FunctionArg(number_t, "value"), ), number_t, math_tan)
))


math_constants = {"math.pi": Decimal(math.pi),
                  "math.e": Decimal(math.e)}



from decimal import Decimal

from pxp.function import FunctionArg, FunctionList, InjectedFunction
from pxp.stdlib.types import number_t, string_t, boolean_t


def str_endswith(resolver, subject, value):
  """Returns True if subject ends with value."""
  sval, val = resolver.resolve_all(subject, value)
  return sval.endswith(val)


def str_find(resolver, subject, value, start, end):
  """Returns the 0-based index of value in subject, or -1 if value is not present in subject.

  start and end are indicies within which to perform the search, if present. The bounds are
  [start, end). If start or end are negative, their position is calculated from the end of the
  string.
  """
  sval, val = resolver.resolve_all(subject, value)
  tval = resolver.resolve(start, none_ok=True)
  nval = resolver.resolve(end, none_ok=True)

  return Decimal(sval.find(val, tval, nval))


def str_len(resolver, subject):
  """Returns the number of characters present in subject."""
  sval = resolver.resolve(subject)
  return Decimal(len(sval))


def str_lower(resolver, subject):
  """Returns subject with all uppercase characters converted to their lowercase equivalents."""
  sval = resolver.resolve(subject)
  return sval.lower()


def str_replace(resolver, subject, old, new, count):
  """Searches through subject for all occurrences of old and replaces them with new. If count is
  given, only at most count replacements will be made.
  """
  sval, oval, nval, cval = resolver.resolve_all(subject, old, new, count)
  return sval.replace(oval, nval, int(cval))


def str_slice(resolver, subject, start, end):
  """Returns a slice (or substring) of subject.

  The bounds are [start, end). start and end can be negative, in which case their position is
  calculated from the end of the string.
  """
  sval = resolver.resolve(subject)
  tval = resolver.resolve(start, none_ok=True)
  nval = resolver.resolve(end, none_ok=True)

  if tval is not None:
    tval = int(tval)
  if nval is not None:
    nval = int(nval)
  return sval[tval:nval]


def str_startswith(resolver, subject, value):
  """Returns True if subject starts with value."""
  sval, val = resolver.resolve_all(subject, value)
  return sval.startswith(val)


def str_strip(resolver, subject):
  """Returns subject with all leading and trailing whitespace removed."""
  sval = resolver.resolve(subject)
  return sval.strip()


def str_upper(resolver, subject):
  """Returns subject will all lowercase characters converted to their uppercase equivalents."""
  sval = resolver.resolve(subject)
  return sval.upper()


string_functions = FunctionList((
  InjectedFunction("str.endswith",
                   (FunctionArg(string_t, "subject"),
                    FunctionArg(string_t, "value")),
                   boolean_t,
                   str_endswith),
  InjectedFunction("str.find",
                   (FunctionArg(string_t, "subject"),
                    FunctionArg(string_t, "value"),
                    FunctionArg(number_t, "start", None),
                    FunctionArg(number_t, "end", None)),
                   number_t,
                   str_find),
  InjectedFunction("str.len", (FunctionArg(string_t, "subject"), ), number_t, str_len),
  InjectedFunction("str.lower", (FunctionArg(string_t, "subject"), ), string_t, str_lower),
  InjectedFunction("str.replace",
                   (FunctionArg(string_t, "subject"),
                    FunctionArg(string_t, "old"),
                    FunctionArg(string_t, "new"),
                    FunctionArg(number_t, "count", Decimal(-1))),
                   string_t,
                   str_replace),
  InjectedFunction("str.slice",
                   (FunctionArg(string_t, "subject"),
                    FunctionArg(number_t, "start", None),
                    FunctionArg(number_t, "end", None)),
                   string_t,
                   str_slice),
  InjectedFunction("str.startswith",
                   (FunctionArg(string_t, "subject"),
                    FunctionArg(string_t, "value")),
                   boolean_t,
                   str_startswith),
  InjectedFunction("str.strip", (FunctionArg(string_t, "subject"), ), string_t, str_strip),
  InjectedFunction("str.upper", (FunctionArg(string_t, "subject"), ), string_t, str_upper)
))



from decimal import Decimal

from pxp.exception import OperatorError
from pxp.function import FunctionArg, FunctionList, InjectedFunction
from pxp.stdlib.types import number_t, string_t, boolean_t


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Number Operators
def op_number_add(resolver, left, right):
  """Returns the sum of two numbers."""
  lval = resolver.resolve(left)
  rval = resolver.resolve(right)
  return lval + rval


def op_number_subtract(resolver, left, right):
  """Returns the difference of two numbers."""
  lval = resolver.resolve(left)
  rval = resolver.resolve(right)
  return lval - rval


def op_number_multiply(resolver, left, right):
  """Returns the product of two numbers."""
  lval = resolver.resolve(left)
  rval = resolver.resolve(right)
  return lval * rval


def op_number_divide(resolver, left, right):
  """Returns the quotient of two numbers."""
  rval = resolver.resolve(right)

  if rval == Decimal(0):
    raise OperatorError("Divide by 0")

  lval = resolver.resolve(left)
  return lval / rval


def op_number_modulus(resolver, left, right):
  """Returns the remainder from left / right."""
  rval = resolver.resolve(right)

  if rval == Decimal(0):
    raise OperatorError("Divide by 0")

  lval = resolver.resolve(left)
  return lval % rval


def op_number_exponentiate(resolver, left, right):
  """Returns the value of left raised to right."""
  lval = resolver.resolve(left)
  rval = resolver.resolve(right)
  return lval ** rval


def op_number_negate(resolver, arg):
  """Returns the negation of arg."""
  aval = resolver.resolve(arg)
  return -1 * aval


def op_number_null_coalesce(resolver, left, right):
  """Returns the left if left is not null, otherwise right. Right is not resolved until it is
  determined that left is null.
  """
  lval = resolver.resolve(left, none_ok=True)
  if lval is not None:
    return lval
  else:
    rval = resolver.resolve(right, none_ok=True)
    return rval


def op_number_cmp_equal(resolver, left, right):
  """Returns True if left is equal to right."""
  lval = resolver.resolve(left)
  rval = resolver.resolve(right)
  return lval == rval


def op_number_cmp_not_equal(resolver, left, right):
  """Returns True if left is not equal to right."""
  lval = resolver.resolve(left)
  rval = resolver.resolve(right)
  return lval != rval


def op_number_cmp_greater_than_or_equal(resolver, left, right):
  """Returns True if left is greater than or equal to right."""
  lval = resolver.resolve(left)
  rval = resolver.resolve(right)
  return lval >= rval


def op_number_cmp_less_than_or_equal(resolver, left, right):
  """Returns True if left is less than or equal to right."""
  lval = resolver.resolve(left)
  rval = resolver.resolve(right)
  return lval <= rval


def op_number_cmp_greater_than(resolver, left, right):
  """Returns True if left is strictly greater than right."""
  lval = resolver.resolve(left)
  rval = resolver.resolve(right)
  return lval > rval


def op_number_cmp_less_than(resolver, left, right):
  """Returns True if left is strictly less than right."""
  lval = resolver.resolve(left)
  rval = resolver.resolve(right)
  return lval < rval


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# String operators
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
def op_string_add(resolver, left, right):
  """Returns the concatenation of left and right."""
  lval = resolver.resolve(left)
  rval = resolver.resolve(right)
  return lval + rval


def op_string_null_coalesce(resolver, left, right):
  """Returns the left if left is not null, otherwise right. Right is not resolved until it is
  determined that left is null.
  """
  lval = resolver.resolve(left, none_ok=True)
  if lval is not None:
    return lval
  else:
    rval = resolver.resolve(right, none_ok=True)
    return rval


def op_string_cmp_equal(resolver, left, right):
  """Returns True if left is lexicographically equal to right."""
  lval = resolver.resolve(left)
  rval = resolver.resolve(right)
  return lval == rval


def op_string_cmp_not_equal(resolver, left, right):
  """Returns True if left is not lexicographically equal to right."""
  lval = resolver.resolve(left)
  rval = resolver.resolve(right)
  return lval != rval


def op_string_cmp_greater_than_or_equal(resolver, left, right):
  """Returns True if left is lexicographically greater than or equal to right."""
  lval = resolver.resolve(left)
  rval = resolver.resolve(right)
  return lval >= rval


def op_string_cmp_less_than_or_equal(resolver, left, right):
  """Returns True if left is lexicographically less than or equal to right."""
  lval = resolver.resolve(left)
  rval = resolver.resolve(right)
  return lval <= rval


def op_string_cmp_greater_than(resolver, left, right):
  """Returns True if left is lexicographically strictly greater than right."""
  lval = resolver.resolve(left)
  rval = resolver.resolve(right)
  return lval > rval


def op_string_cmp_less_than(resolver, left, right):
  """Returns True if left is lexicographically strictly less than right."""
  lval = resolver.resolve(left)
  rval = resolver.resolve(right)
  return lval < rval


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Boolean Operators
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
def op_boolean_null_coalesce(resolver, left, right):
  """Returns the left if left is not null, otherwise right. Right is not resolved until it is
  determined that left is null.
  """
  lval = resolver.resolve(left, none_ok=True)
  if lval is not None:
    return lval
  else:
    rval = resolver.resolve(right, none_ok=True)
    return rval


def op_boolean_cmp_equal(resolver, left, right):
  """Returns True if left is equal to right."""
  lval = resolver.resolve(left)
  rval = resolver.resolve(right)
  return lval == rval


def op_boolean_cmp_not_equal(resolver, left, right):
  """Returns True if left is not equal to right."""
  lval = resolver.resolve(left)
  rval = resolver.resolve(right)
  return lval != rval


def op_boolean_logical_not(resolver, arg):
  """Returns the negation of arg. If arg is True, False is returned. If arg is False, True is
  returned.
  """
  aval = resolver.resolve(arg)
  return not aval


def op_boolean_logical_and(resolver, left, right):
  """Returns True if both left and right evaluate to True, False otherwise.

  If left is not True, the value of right doesn't matter, so right will not be evaluated.
  """
  # Short circuit
  lval = resolver.resolve(left)
  if not lval:
    return False

  rval = resolver.resolve(right)
  if rval:
    return True

  return False


def op_boolean_logical_or(resolver, left, right):
  """Returns True if left or right evaluate to True, False otherwise.

  If left is True, the value of right doesn't matter, so right will not be evaluated.
  """
  # Short circuit
  lval = resolver.resolve(left)
  if lval:
    return True

  rval = resolver.resolve(right)
  if rval:
    return True

  return False


operator_functions = FunctionList((
  InjectedFunction("operator+", (FunctionArg(number_t, "left"), FunctionArg(number_t, "right")), number_t, op_number_add),
  InjectedFunction("operator-", (FunctionArg(number_t, "left"), FunctionArg(number_t, "right")), number_t, op_number_subtract),
  InjectedFunction("operator*", (FunctionArg(number_t, "left"), FunctionArg(number_t, "right")), number_t, op_number_multiply),
  InjectedFunction("operator/", (FunctionArg(number_t, "left"), FunctionArg(number_t, "right")), number_t, op_number_divide),
  InjectedFunction("operator%", (FunctionArg(number_t, "left"), FunctionArg(number_t, "right")), number_t, op_number_modulus),
  InjectedFunction("operator^", (FunctionArg(number_t, "left"), FunctionArg(number_t, "right")), number_t, op_number_exponentiate),
  InjectedFunction("operatorunary-", (FunctionArg(number_t, "arg"), ), number_t, op_number_negate),
  InjectedFunction("operator?", (FunctionArg(number_t, "left"), FunctionArg(number_t, "right")), number_t, op_number_null_coalesce),
  InjectedFunction("operator=", (FunctionArg(number_t, "left"), FunctionArg(number_t, "right")), boolean_t, op_number_cmp_equal),
  InjectedFunction("operator!=", (FunctionArg(number_t, "left"), FunctionArg(number_t, "right")), boolean_t, op_number_cmp_not_equal),
  InjectedFunction("operator>=", (FunctionArg(number_t, "left"), FunctionArg(number_t, "right")), boolean_t, op_number_cmp_greater_than_or_equal),
  InjectedFunction("operator<=", (FunctionArg(number_t, "left"), FunctionArg(number_t, "right")), boolean_t, op_number_cmp_less_than_or_equal),
  InjectedFunction("operator>", (FunctionArg(number_t, "left"), FunctionArg(number_t, "right")), boolean_t, op_number_cmp_greater_than),
  InjectedFunction("operator<", (FunctionArg(number_t, "left"), FunctionArg(number_t, "right")), boolean_t, op_number_cmp_less_than),

  InjectedFunction("operator+", (FunctionArg(string_t, "left"), FunctionArg(string_t, "right")), string_t, op_string_add),
  InjectedFunction("operator?", (FunctionArg(string_t, "left"), FunctionArg(string_t, "right")), string_t, op_string_null_coalesce),
  InjectedFunction("operator=", (FunctionArg(string_t, "left"), FunctionArg(string_t, "right")), boolean_t, op_string_cmp_equal),
  InjectedFunction("operator!=", (FunctionArg(string_t, "left"), FunctionArg(string_t, "right")), boolean_t, op_string_cmp_not_equal),
  InjectedFunction("operator>=", (FunctionArg(string_t, "left"), FunctionArg(string_t, "right")), boolean_t, op_string_cmp_greater_than_or_equal),
  InjectedFunction("operator<=", (FunctionArg(string_t, "left"), FunctionArg(string_t, "right")), boolean_t, op_string_cmp_less_than_or_equal),
  InjectedFunction("operator>", (FunctionArg(string_t, "left"), FunctionArg(string_t, "right")), boolean_t, op_string_cmp_greater_than),
  InjectedFunction("operator<", (FunctionArg(string_t, "left"), FunctionArg(string_t, "right")), boolean_t, op_string_cmp_less_than),

  InjectedFunction("operator?", (FunctionArg(boolean_t, "left"), FunctionArg(boolean_t, "right")), boolean_t, op_boolean_null_coalesce),
  InjectedFunction("operator=", (FunctionArg(boolean_t, "left"), FunctionArg(boolean_t, "right")), boolean_t, op_boolean_cmp_equal),
  InjectedFunction("operator!=", (FunctionArg(boolean_t, "left"), FunctionArg(boolean_t, "right")), boolean_t, op_boolean_cmp_not_equal),
  InjectedFunction("operatorunary!", (FunctionArg(boolean_t, "arg"), ), boolean_t, op_boolean_logical_not),
  InjectedFunction("operator&", (FunctionArg(boolean_t, "left"), FunctionArg(boolean_t, "right")), boolean_t, op_boolean_logical_and),
  InjectedFunction("operator|", (FunctionArg(boolean_t, "left"), FunctionArg(boolean_t, "right")), boolean_t, op_boolean_logical_or)
))



"""This module contains general pxp standard library functions."""

from decimal import Decimal, InvalidOperation

from pxp.exception import FunctionError
from pxp.function import FunctionArg, FunctionList, InjectedFunction
from pxp.stdlib.types import number_t, string_t, boolean_t


def branch_if(resolver, predicate, if_true, if_false):
  """Returns if_true if predicate evaluates to true, if_false otherwise.

  predicate is resolved and evaluated first, and then only the argument that will be returned is
  resolved and evaluated.
  """
  pval = resolver.resolve(predicate)
  if predicate:
    return resolver.resolve(if_true)
  else:
    return resolver.resolve(if_false)


def is_number(resolver, value):
  """Returns True if the string value can be parsed as a number."""
  try:
    Decimal(resolver.resolve(value))
    return True
  except InvalidOperation:
    return False


def to_number(resolver, value):
  """Returns the number representation of the string value.

  This is an unchecked conversion, so if the string is not a valid number an exception will be
  thrown.
  """
  try:
    return Decimal(resolver.resolve(value))
  except InvalidOperation:
    raise FunctionError("Unable to parse string as a number")


def to_string(resolver, value):
  """Returns the string representation of the value."""
  return str(resolver.resolve(value))


def is_null(resolver, value):
  """Returns True if the value is None."""
  return resolver.resolve(value, none_ok=True) is None


general_functions = FunctionList((
  InjectedFunction("if",
                   (FunctionArg(boolean_t, "predicate"),
                    FunctionArg(number_t, "if_true"),
                    FunctionArg(number_t, "if_false")),
                   number_t,
                   branch_if),
  InjectedFunction("if",
                   (FunctionArg(boolean_t, "predicate"),
                    FunctionArg(string_t, "if_true"),
                    FunctionArg(string_t, "if_false")),
                   string_t,
                   branch_if),
  InjectedFunction("if",
                   (FunctionArg(boolean_t, "predicate"),
                    FunctionArg(boolean_t, "if_true"),
                    FunctionArg(boolean_t, "if_false")),
                   boolean_t,
                   branch_if),
  InjectedFunction("is_num", (FunctionArg(string_t, "value"), ), boolean_t, is_number),
  InjectedFunction("to_num", (FunctionArg(string_t, "value"), ), number_t, to_number),
  InjectedFunction("to_str", (FunctionArg(number_t, "value"), ), string_t, to_string),
  InjectedFunction("to_str", (FunctionArg(string_t, "value"), ), string_t, to_string),
  InjectedFunction("to_str", (FunctionArg(boolean_t, "value"), ), string_t, to_string),
  InjectedFunction("is_null", (FunctionArg(number_t, "value"), ), boolean_t, is_null),
  InjectedFunction("is_null", (FunctionArg(string_t, "value"), ), boolean_t, is_null),
  InjectedFunction("is_null", (FunctionArg(boolean_t, "value"), ), boolean_t, is_null)
))



class PXPError(Exception):
  """Base class for all PXPErrors."""
  pass


class ScopeError(PXPError):
  """Error type thrown by pxp.source.SourceBase classes."""
  pass


class PositionalError(PXPError):
  """Position error type. Contains a source-code position accompanying the error."""
  def __init__(self, position, *args):
    """Initializes the PositionalError instance."""
    super().__init__(*args)
    self.position = position

  def __str__(self):
    """The string representation of the error with the message and the 1-based coordinates of the
    position.
    """
    return "{0} at position {1}".format(self.message, self.position1)

  @property
  def message(self):
    """The error message."""
    return super().__str__()

  @property
  def position1(self):
    """The 1-based position of the offending code."""
    line, char = self.position
    return (line + 1, char + 1)


class CompilerError(PositionalError):
  """Error type thrown by the Compiler."""
  pass


class RuntimeError(PositionalError):
  """Error type thrown by the runtime (interpreter)."""
  pass


class FunctionError(PXPError):
  """Error type thrown by functions."""
  pass


class OperatorError(FunctionError):
  """Error type thrown by operators."""
  pass



from setuptools import setup, find_packages

version = "0.2.6"

setup(name="pxp",
      version=version,
      license="BSD",
      platforms="any",
      author="Trey Cucco",
      author_email="fcucco@gmail.com",
      packages=find_packages(exclude=["config", "tests"]),
      description="A simple python-hosted expression language.",
      url="https://github.com/treycucco/pxp",
      download_url="https://github.com/treycucco/pxp/tarball/master",
      classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.4"
      ],
      install_requires=["pyebnf>=0.0.6"])



from enum import Enum

from pyebnf.parser_base import ParserBase, get_printable_character, get_whitespace_character
from pyebnf.primitive import Node, NodeType, alternation, concatenation
from pyebnf.primitive import exclusion, option, repetition, terminal




class TokenType(Enum):
  program = 1
  statement = 2
  assignment = 3
  variable = 4
  identifier = 5
  simple_identifier = 6
  complex_identifier = 7
  alpha_character = 8
  alpha_character_upper = 9
  alpha_character_lower = 10
  digit = 11
  reserved_words = 12
  expression = 13
  expression_terminal = 14
  number = 15
  string = 16
  double_string_char = 17
  single_string_char = 18
  all_characters = 19
  function_call = 20
  function_name = 21
  function_args = 22
  subexpression = 23
  branch = 24
  branch_if = 25
  branch_elif = 26
  branch_else = 27
  switch = 28
  switch_subject = 29
  switch_case = 30
  switch_default = 31
  operator = 32
  directive = 33
  eol = 34
  whitespace = 35
  comment = 36
  return_statement = 37
  eof = 38


class Parser(ParserBase):
  def entry_point(self, text):
    yield from self._program(text)

  def _gen_program(self, gen, text):
    yield from Node.iterate_merge(TokenType.program, gen, text)

  def _gen_statement(self, gen, text):
    yield from gen(text)

  def _gen_assignment(self, gen, text):
    yield from Node.iterate_merge(TokenType.assignment, gen, text)

  def _gen_variable(self, gen, text):
    yield from Node.iterate_reduce(TokenType.variable, gen, text)

  def _gen_identifier(self, gen, text):
    yield from gen(text)

  def _gen_simple_identifier(self, gen, text):
    yield from gen(text)

  def _gen_complex_identifier(self, gen, text):
    yield from gen(text)

  def _gen_alpha_character(self, gen, text):
    yield from Node.iterate_reduce(TokenType.alpha_character, gen, text)

  def _gen_alpha_character_upper(self, gen, text):
    yield from gen(text)

  def _gen_alpha_character_lower(self, gen, text):
    yield from gen(text)

  def _gen_digit(self, gen, text):
    yield from gen(text)

  def _gen_reserved_words(self, gen, text):
    yield from gen(text)

  def _gen_expression(self, gen, text):
    yield from Node.iterate_merge(TokenType.expression, gen, text)

  def _gen_expression_terminal(self, gen, text):
    yield from gen(text)

  def _gen_number(self, gen, text):
    yield from Node.iterate_reduce(TokenType.number, gen, text)

  def _gen_string(self, gen, text):
    yield from Node.iterate_reduce(TokenType.string, gen, text)

  def _gen_double_string_char(self, gen, text):
    yield from gen(text)

  def _gen_single_string_char(self, gen, text):
    yield from gen(text)

  def _gen_all_characters(self, gen, text):
    yield from gen(text)

  def _gen_function_call(self, gen, text):
    yield from Node.iterate_merge(TokenType.function_call, gen, text)

  def _gen_function_name(self, gen, text):
    yield from Node.iterate_reduce(TokenType.function_name, gen, text)

  def _gen_function_args(self, gen, text):
    yield from Node.iterate_merge(TokenType.function_args, gen, text)

  def _gen_subexpression(self, gen, text):
    yield from Node.iterate_merge(TokenType.subexpression, gen, text)

  def _gen_branch(self, gen, text):
    yield from Node.iterate_merge(TokenType.branch, gen, text)

  def _gen_branch_if(self, gen, text):
    yield from Node.iterate_merge(TokenType.branch_if, gen, text)

  def _gen_branch_elif(self, gen, text):
    yield from Node.iterate_merge(TokenType.branch_elif, gen, text)

  def _gen_branch_else(self, gen, text):
    yield from Node.iterate_merge(TokenType.branch_else, gen, text)

  def _gen_switch(self, gen, text):
    yield from Node.iterate_merge(TokenType.switch, gen, text)

  def _gen_switch_subject(self, gen, text):
    yield from Node.iterate_merge(TokenType.switch_subject, gen, text)

  def _gen_switch_case(self, gen, text):
    yield from Node.iterate_merge(TokenType.switch_case, gen, text)

  def _gen_switch_default(self, gen, text):
    yield from Node.iterate_merge(TokenType.switch_default, gen, text)

  def _gen_operator(self, gen, text):
    yield from Node.iterate_reduce(TokenType.operator, gen, text)

  def _gen_directive(self, gen, text):
    yield from Node.iterate_reduce(TokenType.directive, gen, text)

  def _gen_eol(self, gen, text):
    yield from gen(text)

  def _gen_whitespace(self, gen, text):
    yield from gen(text)

  def _gen_comment(self, gen, text):
    yield from Node.iterate_reduce(TokenType.comment, gen, text)

  def _gen_return_statement(self, gen, text):
    yield from Node.iterate_merge(TokenType.return_statement, gen, text)

  def _gen_eof(self, gen, text):
    yield from gen(text)

  def _program(self, text):
    """program = {directive} , {statement} , return_statement , "\eof" ;"""
    self.attempting(text)
    gen = concatenation([
      repetition(
        self._directive,
        bound=-1
      ),
      repetition(
        self._statement,
        bound=-1
      ),
      self._return_statement,
      terminal("\\eof")
    ], True)
    yield from self._gen_program(gen, text)

  def _statement(self, text):
    """statement = comment | assignment , ";" ;"""
    self.attempting(text)
    gen = alternation([
      self._comment,
      concatenation([
        self._assignment,
        terminal(";")
      ], True)
    ])
    yield from self._gen_statement(gen, text)

  def _assignment(self, text):
    """assignment = variable , (":=" | "<-") , expression ;"""
    self.attempting(text)
    gen = concatenation([
      self._variable,
      alternation([
        terminal(":="),
        terminal("<-")
      ]),
      self._expression
    ], True)
    yield from self._gen_assignment(gen, text)

  def _variable(self, text):
    """variable = identifier - reserved_words ;"""
    self.attempting(text)
    gen = exclusion(
      self._identifier,
      self._reserved_words
    )
    yield from self._gen_variable(gen, text)

  def _identifier(self, text):
    """identifier = simple_identifier | complex_identifier ;"""
    self.attempting(text)
    gen = alternation([
      self._simple_identifier,
      self._complex_identifier
    ])
    yield from self._gen_identifier(gen, text)

  def _simple_identifier(self, text):
    """simple_identifier = (alpha_character | "_") . {alpha_character | digit | "_" | "."} ;"""
    self.attempting(text)
    gen = concatenation([
      alternation([
        self._alpha_character,
        terminal("_")
      ]),
      repetition(
        alternation([
          self._alpha_character,
          self._digit,
          terminal("_"),
          terminal(".")
        ]),
        bound=-1
      )
    ], False)
    yield from self._gen_simple_identifier(gen, text)

  def _complex_identifier(self, text):
    """complex_identifier = "[" . all_characters - "]" . {all_characters - "]"} . "]" ;"""
    self.attempting(text)
    gen = concatenation([
      terminal("["),
      exclusion(
        self._all_characters,
        terminal("]")
      ),
      repetition(
        exclusion(
          self._all_characters,
          terminal("]")
        ),
        bound=-1
      ),
      terminal("]")
    ], False)
    yield from self._gen_complex_identifier(gen, text)

  def _alpha_character(self, text):
    """alpha_character = alpha_character_upper | alpha_character_lower ;"""
    self.attempting(text)
    gen = alternation([
      self._alpha_character_upper,
      self._alpha_character_lower
    ])
    yield from self._gen_alpha_character(gen, text)

  def _alpha_character_upper(self, text):
    """alpha_character_upper = "A" | "B" | "C" | "D" | "E" | "F" | "G" | "H" | "I"
                             | "J" | "K" | "L" | "M" | "N" | "O" | "P" | "Q" | "R"
                             | "S" | "T" | "U" | "V" | "W" | "X" | "Y" | "Z" ;"""
    self.attempting(text)
    gen = alternation([
      terminal("A"),
      terminal("B"),
      terminal("C"),
      terminal("D"),
      terminal("E"),
      terminal("F"),
      terminal("G"),
      terminal("H"),
      terminal("I"),
      terminal("J"),
      terminal("K"),
      terminal("L"),
      terminal("M"),
      terminal("N"),
      terminal("O"),
      terminal("P"),
      terminal("Q"),
      terminal("R"),
      terminal("S"),
      terminal("T"),
      terminal("U"),
      terminal("V"),
      terminal("W"),
      terminal("X"),
      terminal("Y"),
      terminal("Z")
    ])
    yield from self._gen_alpha_character_upper(gen, text)

  def _alpha_character_lower(self, text):
    """alpha_character_lower = "a" | "b" | "c" | "d" | "e" | "f" | "g" | "h" | "i"
                             | "j" | "k" | "l" | "m" | "n" | "o" | "p" | "q" | "r"
                             | "s" | "t" | "u" | "v" | "w" | "x" | "y" | "z" ;"""
    self.attempting(text)
    gen = alternation([
      terminal("a"),
      terminal("b"),
      terminal("c"),
      terminal("d"),
      terminal("e"),
      terminal("f"),
      terminal("g"),
      terminal("h"),
      terminal("i"),
      terminal("j"),
      terminal("k"),
      terminal("l"),
      terminal("m"),
      terminal("n"),
      terminal("o"),
      terminal("p"),
      terminal("q"),
      terminal("r"),
      terminal("s"),
      terminal("t"),
      terminal("u"),
      terminal("v"),
      terminal("w"),
      terminal("x"),
      terminal("y"),
      terminal("z")
    ])
    yield from self._gen_alpha_character_lower(gen, text)

  def _digit(self, text):
    """digit = "0" | "1" | "2" | "3" | "4" | "5" | "6" | "7" | "8" | "9" ;"""
    self.attempting(text)
    gen = alternation([
      terminal("0"),
      terminal("1"),
      terminal("2"),
      terminal("3"),
      terminal("4"),
      terminal("5"),
      terminal("6"),
      terminal("7"),
      terminal("8"),
      terminal("9")
    ])
    yield from self._gen_digit(gen, text)

  def _reserved_words(self, text):
    """reserved_words = "return" | "if" | "elif" | "else" | "switch" | "when";"""
    self.attempting(text)
    gen = alternation([
      terminal("return"),
      terminal("if"),
      terminal("elif"),
      terminal("else"),
      terminal("switch"),
      terminal("when")
    ])
    yield from self._gen_reserved_words(gen, text)

  def _expression(self, text):
    """expression = {operator} , expression_terminal , {operator , {operator} , expression} ;"""
    self.attempting(text)
    gen = concatenation([
      repetition(
        self._operator,
        bound=-1
      ),
      self._expression_terminal,
      repetition(
        concatenation([
          self._operator,
          repetition(
            self._operator,
            bound=-1
          ),
          self._expression
        ], True),
        bound=-1
      )
    ], True)
    yield from self._gen_expression(gen, text)

  def _expression_terminal(self, text):
    """expression_terminal = function_call
                           | number
                           | string
                           | variable
                           | subexpression
                           | branch
                           | switch ;"""
    self.attempting(text)
    gen = alternation([
      self._function_call,
      self._number,
      self._string,
      self._variable,
      self._subexpression,
      self._branch,
      self._switch
    ])
    yield from self._gen_expression_terminal(gen, text)

  def _number(self, text):
    """number = ["-"] . ("0" | digit - "0" . {digit}) . ["." . digit . {digit}] ;"""
    self.attempting(text)
    gen = concatenation([
      option(
        terminal("-")
      ),
      alternation([
        terminal("0"),
        concatenation([
          exclusion(
            self._digit,
            terminal("0")
          ),
          repetition(
            self._digit,
            bound=-1
          )
        ], False)
      ]),
      option(
        concatenation([
          terminal("."),
          self._digit,
          repetition(
            self._digit,
            bound=-1
          )
        ], False)
      )
    ], False)
    yield from self._gen_number(gen, text)

  def _string(self, text):
    """string = '"' . {double_string_char} . '"'
              | "'" . {single_string_char} . "'" ;"""
    self.attempting(text)
    gen = alternation([
      concatenation([
        terminal('"'),
        repetition(
          self._double_string_char,
          bound=-1
        ),
        terminal('"')
      ], False),
      concatenation([
        terminal("'"),
        repetition(
          self._single_string_char,
          bound=-1
        ),
        terminal("'")
      ], False)
    ])
    yield from self._gen_string(gen, text)

  def _double_string_char(self, text):
    """double_string_char = "\\" , all_characters - '"'
                          | "\" , '"'
                          | all_characters - '"' ;"""
    self.attempting(text)
    gen = alternation([
      concatenation([
        terminal("\\\\"),
        exclusion(
          self._all_characters,
          terminal('"')
        )
      ], True),
      concatenation([
        terminal("\\"),
        terminal('"')
      ], True),
      exclusion(
        self._all_characters,
        terminal('"')
      )
    ])
    yield from self._gen_double_string_char(gen, text)

  def _single_string_char(self, text):
    """single_string_char = "\\" , all_characters - "'"
                          | "\" , "'"
                          | all_characters - "'" ;"""
    self.attempting(text)
    gen = alternation([
      concatenation([
        terminal("\\\\"),
        exclusion(
          self._all_characters,
          terminal("'")
        )
      ], True),
      concatenation([
        terminal("\\"),
        terminal("'")
      ], True),
      exclusion(
        self._all_characters,
        terminal("'")
      )
    ])
    yield from self._gen_single_string_char(gen, text)

  def _all_characters(self, text):
    """all_characters = ? all_printable_characters ? ;"""
    self.attempting(text)
    gen = get_printable_character
    yield from self._gen_all_characters(gen, text)

  def _function_call(self, text):
    """function_call = function_name . "(" , [function_args] , ")" ;"""
    self.attempting(text)
    gen = concatenation([
      self._function_name,
      concatenation([
        terminal("("),
        option(
          self._function_args
        ),
        terminal(")")
      ], True)
    ], False)
    yield from self._gen_function_call(gen, text)

  def _function_name(self, text):
    """function_name = identifier ;"""
    self.attempting(text)
    gen = self._identifier
    yield from self._gen_function_name(gen, text)

  def _function_args(self, text):
    """function_args = expression , {"," , function_args} ;"""
    self.attempting(text)
    gen = concatenation([
      self._expression,
      repetition(
        concatenation([
          terminal(","),
          self._function_args
        ], True),
        bound=-1
      )
    ], True)
    yield from self._gen_function_args(gen, text)

  def _subexpression(self, text):
    """subexpression = "(" , expression , ")" ;"""
    self.attempting(text)
    gen = concatenation([
      terminal("("),
      self._expression,
      terminal(")")
    ], True)
    yield from self._gen_subexpression(gen, text)

  def _branch(self, text):
    """branch = branch_if , {branch_elif} , branch_else ;"""
    self.attempting(text)
    gen = concatenation([
      self._branch_if,
      repetition(
        self._branch_elif,
        bound=-1
      ),
      self._branch_else
    ], True)
    yield from self._gen_branch(gen, text)

  def _branch_if(self, text):
    """branch_if = "if" , expression , ":" , expression ;"""
    self.attempting(text)
    gen = concatenation([
      terminal("if"),
      self._expression,
      terminal(":"),
      self._expression
    ], True)
    yield from self._gen_branch_if(gen, text)

  def _branch_elif(self, text):
    """branch_elif = "elif" , expression , ":" , expression ;"""
    self.attempting(text)
    gen = concatenation([
      terminal("elif"),
      self._expression,
      terminal(":"),
      self._expression
    ], True)
    yield from self._gen_branch_elif(gen, text)

  def _branch_else(self, text):
    """branch_else = "else" , ":" , expression ;"""
    self.attempting(text)
    gen = concatenation([
      terminal("else"),
      terminal(":"),
      self._expression
    ], True)
    yield from self._gen_branch_else(gen, text)

  def _switch(self, text):
    """switch = switch_subject , switch_case , {switch_case} , [switch_default] ;"""
    self.attempting(text)
    gen = concatenation([
      self._switch_subject,
      self._switch_case,
      repetition(
        self._switch_case,
        bound=-1
      ),
      option(
        self._switch_default
      )
    ], True)
    yield from self._gen_switch(gen, text)

  def _switch_subject(self, text):
    """switch_subject = "switch" , expression ;"""
    self.attempting(text)
    gen = concatenation([
      terminal("switch"),
      self._expression
    ], True)
    yield from self._gen_switch_subject(gen, text)

  def _switch_case(self, text):
    """switch_case = "when" , expression , ":" , expression ;"""
    self.attempting(text)
    gen = concatenation([
      terminal("when"),
      self._expression,
      terminal(":"),
      self._expression
    ], True)
    yield from self._gen_switch_case(gen, text)

  def _switch_default(self, text):
    """switch_default = "else" , ":" , expression ;"""
    self.attempting(text)
    gen = concatenation([
      terminal("else"),
      terminal(":"),
      self._expression
    ], True)
    yield from self._gen_switch_default(gen, text)

  def _operator(self, text):
    """operator = "+" | "-"  | "*"  | "/"  | "%" | "^"
                | "=" | "!=" | ">=" | "<=" | ">" | "<"
                | "!" | "|"  | "&"  | "?" ;"""
    self.attempting(text)
    gen = alternation([
      terminal("+"),
      terminal("-"),
      terminal("*"),
      terminal("/"),
      terminal("%"),
      terminal("^"),
      terminal("="),
      terminal("!="),
      terminal(">="),
      terminal("<="),
      terminal(">"),
      terminal("<"),
      terminal("!"),
      terminal("|"),
      terminal("&"),
      terminal("?")
    ])
    yield from self._gen_operator(gen, text)

  def _directive(self, text):
    """directive = "#" . { all_characters - "\n" } . eol ;"""
    self.attempting(text)
    gen = concatenation([
      terminal("#"),
      repetition(
        exclusion(
          self._all_characters,
          terminal("\n")
        ),
        bound=-1
      ),
      self._eol
    ], False)
    yield from self._gen_directive(gen, text)

  def _eol(self, text):
    """eol = { whitespace - "\n" } . "\n" ;"""
    self.attempting(text)
    gen = concatenation([
      repetition(
        exclusion(
          self._whitespace,
          terminal("\n")
        ),
        bound=-1
      ),
      terminal("\n")
    ], False)
    yield from self._gen_eol(gen, text)

  def _whitespace(self, text):
    """whitespace = ? all_whitespace_characters ? ;"""
    self.attempting(text)
    gen = get_whitespace_character
    yield from self._gen_whitespace(gen, text)

  def _comment(self, text):
    """comment = "/*" . {all_characters - "*" | "*" . all_characters - "/"} . "*/" ;"""
    self.attempting(text)
    gen = concatenation([
      terminal("/*"),
      repetition(
        alternation([
          exclusion(
            self._all_characters,
            terminal("*")
          ),
          concatenation([
            terminal("*"),
            exclusion(
              self._all_characters,
              terminal("/")
            )
          ], False)
        ]),
        bound=-1
      ),
      terminal("*/")
    ], False)
    yield from self._gen_comment(gen, text)

  def _return_statement(self, text):
    """return_statement = ["return"] , expression , [";"] ;"""
    self.attempting(text)
    gen = concatenation([
      option(
        terminal("return")
      ),
      self._expression,
      option(
        terminal(";")
      )
    ], True)
    yield from self._gen_return_statement(gen, text)

  def _eof(self, text):
    """eof = { eol } . eol ;"""
    self.attempting(text)
    gen = concatenation([
      repetition(
        self._eol,
        bound=-1
      ),
      self._eol
    ], False)
    yield from self._gen_eof(gen, text)




"""This module contains classes and methods for working with scopes in PXP programs. Scopes are
collections of variables, constants and functions that programs can refer to.
"""

from contextlib import contextmanager

from pxp.exception import ScopeError
from pxp.function import FunctionList


class ScopeBase(object):
  """The base class for all Scope classes. This class is abstract."""
  def __init__(self):
    """Initialize the ScopeBase instance."""
    pass

  def contains_variable(self, name):
    """Returns True if the scope contains a variable or constant matching the given name."""
    raise NotImplementedError()

  def is_constant(self, name):
    """Returns True if the scope contains a constant matching the given name."""
    raise NotImplementedError()

  def get_variable(self, name):
    """Returns the value registered for a variable or constant in the scope."""
    raise NotImplementedError()

  def set_variable(self, name, value):
    """Sets the value of a variable in the scope."""
    raise NotImplementedError()

  def overwrite_variable(self, name, value):
    """Overwrite the value of a variable in the scope."""
    raise NotImplementedError()

  def contains_function(self, signature):
    """Returns True if the scope contains the function with the given signature."""
    raise NotImplementedError()

  def get_function(self, signature):
    """Returns the Function object registered with the given signature."""
    raise NotImplementedError()


class DictScope(ScopeBase):
  """A subclass of ScopeBase that contains the variables and constants in a python dictionary."""
  def __init__(self, variables=None, constants=None, functions=None):
    """Construct the DictScope instance."""
    super().__init__()
    self.variables = variables or {}
    self.constants = constants or {}
    self.functions = functions or FunctionList()

  def contains_variable(self, name):
    """Returns True if the scope contains a variable or constant matching the given name."""
    return name in self.variables or self.is_constant(name)

  def is_constant(self, name):
    """Returns True if the scope contains a constant matching the given name."""
    return name in self.constants

  def get_variable(self, name):
    """Returns the variable or constant with the given name. If the scope contains both a
    variable and a constant with the given name, the variable is returned. If no variable or
    constant with the given name is found, a ScopeError is raised.
    """
    if name in self.variables:
      return self.variables[name]
    elif self.is_constant(name):
      return self.constants[name]
    else:
      raise ScopeError("Unknown variable or constant {0}".format(name))

  def set_variable(self, name, value):
    """Sets the value of the variable. If a constant is registered with the given name, a
    ScopeError is raised.
    """
    if self.is_constant(name):
      raise ScopeError("Cannot redefine constant {0}".format(name))
    else:
      self.variables[name] = value

  def overwrite_variable(self, name, value):
    """Passes through to set_veriable."""
    self.set_variable(name, value)

  def contains_function(self, signature):
    """Returns True if a function has been registered with the scope with the given signature."""
    return signature in self.functions

  def get_function(self, signature):
    """Returns the function registered with the given signature."""
    return self.functions[signature]


class ScopeStack(ScopeBase):
  """A scope stack is a specialization of ScopeBase that contains a stack of scopes, and will
  search through them from top to bottom looking for variables and functions.
  """
  def __init__(self, *scopes):
    """Initialize the ScopeStack instance."""
    super().__init__()
    self.scopes = list(scopes)

  @property
  def current_scope(self):
    """The current (top-most) scope."""
    return self.scopes[-1]

  def contains_variable(self, name):
    """Searches through all scopes from top to bottom for a variable or constant with the given
    name, returning True if one is found.
    """
    for scope in reversed(self.scopes):
      if scope.contains_variable(name):
        return True
    return False

  def is_constant(self, name):
    """Searches through all scopes from top to bottom for a constant with the given name, returning
    True if one is found."""
    for scope in reversed(self.scopes):
      if scope.contains_variable(name):
        return scope.is_constant(name)
    return False

  def get_variable(self, name):
    """Searches through all scopes from top to bottom and returns the first variable or constant
    found in one of the scopes.
    """
    for scope in reversed(self.scopes):
      if scope.contains_variable(name):
        return scope.get_variable(name)
    raise ScopeError("Undefined variable {0}".format(name))

  def set_variable(self, name, value):
    """Sets the value of the variable in current_scope."""
    self.current_scope.set_variable(name, value)

  def overwrite_variable(self, name, value):
    """Searches through alls copes from top to bottom and finds the first the contains variable, and
    overwrites the value.
    """
    for scope in reversed(self.scopes):
      if scope.contains_variable(name):
        scope.set_variable(name, value)
        break
    else:
      raise ScopeError("Undefined variable {0}".format(name))

  def contains_function(self, signature):
    """Searches through all scopes from top to bottom and for a function matching
    """
    for scope in reversed(self.scopes):
      if scope.contains_function(signature):
        return True
    return False

  def get_function(self, signature):
    """Searches through all scopes from top to bottom and returns the first function found that
    matches the given signature.
    """
    for scope in reversed(self.scopes):
      if scope.contains_function(signature):
        return scope.get_function(signature)
    raise ScopeError("Undefined function {0}".format(signature))

  def push(self, scope):
    """Add a scope to the top of the stack."""
    if isinstance(scope, list):
      self.scopes.extend(scope)
    else:
      self.scopes.append(scope)
    return self

  def pop(self, n=1):
    """Pops the current_scope from the stack."""
    return self.scopes.pop()

  @contextmanager
  def using(self, scope):
    """A context manager that adds a scope, executes the code block, and pops the scope off the
    stack.
    """
    self.push(scope)
    try:
      yield self
    finally:
      self.pop()



"""This module contains classes and methods for defining functions that can be called from PXP
programs.
"""

from collections import namedtuple


_Undefined = object()
FunctionArg = namedtuple("FunctionArg", ["type", "name", "default", "description"])
FunctionArg.__new__.__defaults__ = (_Undefined, None)


class Function(object):
  """This is the base class for implementations of PXP functions."""
  def __init__(self, name, args, return_type):
    """Construct the Function instance."""
    self.name = name
    self.args = args
    self.return_type = return_type
    self._signatures = None

  @property
  def signatures(self):
    """Each function has one or more type signatures.

    A type signature is a tuple containing the name of the function, and type types of its
    arguments.

    If a function has any arguments with default values, it will have multiple signatures. E.G.
    if a function takes two Number arguments, the second of which has a default value of 0, it
    would have two signatures:

    (function_name, "Number", "Number"), and
    (function_name, "Number")
    """
    if self._signatures is None:
      args = list(self.args)
      self._signatures = [(self.name, ) + tuple(a.type for a in args)]

      while args and args[-1].default is not _Undefined:
        args = args[:-1]
        self._signatures.append((self.name, ) + tuple(a.type for a in args))
    return self._signatures

  @property
  def doc(self):
    raise NotImplementedError()

  def call(self):
    """The implementation of the function, __call__ builds the argument list and delegates the call
    to this method."""
    raise NotImplementedError()


  def __call__(self, resolver, *args, **kwargs):
    """Given a resolver and args and keywords args, build an argument list to be passed to the
    call method of the class. The args and keyword args are checked against the order and names of
    the class' args list.
    """
    send_args = []
    for i, arg in enumerate(self.args):
      if i < len(args):
        send_args.append(args[i])
      elif arg.name in kwargs:
        send_args.append(kwargs[arg.name])
      else:
        if arg.default is _Undefined:
          raise Exception("No value given for required "
                          "argument {0} in function {1}".format(arg.name,
                                                                self.name))
        else:
          send_args.append(arg.default)
    return self.call(resolver, *send_args)


class InjectedFunction(Function):
  """A subclass of Function that accepts a function which .call delegates to."""
  def __init__(self, name, args, return_type, call):
    """Constructs the InjectedFunction instance."""
    super().__init__(name, args, return_type)
    self._call = call

  @property
  def doc(self):
    return self._call.__doc__

  def call(self, *args):
    """Pass *args to the function given to the constructor."""
    return self._call(*args)


class FunctionList(object):
  """FunctionList is an object that holds a list of functions, registered by their signatures and
  allows the lookup of functions based on their signature.
  """
  def __init__(self, registrants=None):
    """Constructs the FunctionList instance."""
    self.functions = {}
    if registrants:
      for function in registrants:
        self.register(function)

  def __contains__(self, signature):
    """Returns True if the given signature matches one of the registered functions."""
    return signature in self.functions

  def __getitem__(self, signature):
    """Returns the function registered under signature."""
    return self.functions[signature]

  def __setitem__(self, signature, function):
    """Registers a function with signature."""
    if signature in self:
      raise Exception("Redefinition of function {0}".format(signature))
    else:
      self.functions[signature] = function

  def register(self, function):
    """Register a function. Iterates through all the function's signatures and registers the
    function under each.
    """
    for signature in function.signatures:
      self[signature] = function
    return self

  def merge(self, *others):
    """Merge other function lists into this instance.

    Each argument can be either a FunctionList, a list, or a tuple.

    If the argument is a list or tuple, each item is expected to be a Function instance and the
    functions are passed to register.
    """
    for other in others:
      if isinstance(other, (list, tuple)):
        for fxn in other:
          self.register(fxn)
      elif isinstance(other, FunctionList):
        for sig, fxn in other.functions.items():
          self[sig] = fxn
    return self
